import java.io.BufferedOutputStream;
import java.io.BufferedReader;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.PrintStream;
import java.util.ArrayList;



public class VHDL_Entity 
{
	public String name = null;
	public ArrayList <VHDL_Signal> entitySignals = new ArrayList <VHDL_Signal>();
	public ArrayList <VHDL_Signal> genericSignals = new ArrayList <VHDL_Signal>();

	private String filename = null;
	private boolean parsed = false;
	private ArrayList <String> beforeIOSignals = new ArrayList<String>();
	private ArrayList <String> beforeInternalSignals = new ArrayList<String>();
	private ArrayList <String> beforeStructuralInstances = new ArrayList<String>();
	private ArrayList <String> rest = new ArrayList<String>();
	
	private ArrayList <VHDL_Signal> toAddSignals = new ArrayList <VHDL_Signal>();
	private ArrayList <VHDL_Entity> toAddTypes = new ArrayList <VHDL_Entity>();
	private ArrayList <String> toAddInstanceNames = new ArrayList <String>();
	private ArrayList <String> toAddBehavior = new ArrayList<String>();
	private ArrayList <ArrayList <String> > toAddMappings = new ArrayList <ArrayList <String>>();
	
	// used for parsing IO and generic signals incrementally
	private static final int STATE_NAME = 0;
	private static final int STATE_COLON = 1;
	private static final int STATE_DIRECTION = 2;
	private static final int STATE_TYPE = 3;
	private static final int STATE_INITIAL = 4;
	private int entity_SignalsState = STATE_NAME;
	private VHDL_Signal current_Signal = null;
	
	// types for incremental parsing
	private final static int TYPE_GENERIC = 0;
	private final static int TYPE_IO = 1;
	
	// current stage in parsing a file
	private final static int STAGE_BEFORE_IO = 0;
	private final static int STAGE_BEFORE_INTERNAL = 1;
	private final static int STAGE_BEFORE_STRUCTURAL = 2;
	private final static int STAGE_ENDING = 3;
	
	public VHDL_Entity()
	{
		filename = "blank";
	}
	
	public VHDL_Entity(String filename)
	{
		this.filename = filename;
	}
	
	public boolean createBlank(String entityName)
	{
		boolean rv = true;
		
		if (parsed)
		{
			System.out.println("VHDL_Entity was parsed twice (in createBlank())?!? filename:" + filename);
			rv = false;
		}
		else
		{		
			parsed = true;
			name = entityName;
			
			//private ArrayList <String> beforeIOSignals = new ArrayList<String>();
			beforeIOSignals.add("-- Created from Template Blank Module using VHDL_Entity Editor by Stan Bak");
			beforeIOSignals.add("");
			beforeIOSignals.add("library IEEE;");
			beforeIOSignals.add("use IEEE.STD_LOGIC_1164.ALL;");
			beforeIOSignals.add("use IEEE.STD_LOGIC_ARITH.ALL;");
			beforeIOSignals.add("use IEEE.STD_LOGIC_UNSIGNED.ALL;");
			beforeIOSignals.add("");
			beforeIOSignals.add("entity " + entityName + " is");
			beforeIOSignals.add("\tport ("); 			
			//private ArrayList <String> beforeInternalSignals = new ArrayList<String>();
			
			beforeInternalSignals.add(");"); 
			beforeInternalSignals.add("end " + entityName + ";");
			beforeInternalSignals.add(""); 
			beforeInternalSignals.add("architecture Behavioral of " + entityName + " is");

			beforeStructuralInstances.add("begin");

			rest.add("end Behavioral;");
			rest.add("");				
		}
		
		return rv;
	}
	
	/**
	 * Parse the file
	 * @return true iff there were no errors
	 */
	public boolean parse()
	{
		boolean rv = true;
		
		if (parsed)
		{
			System.out.println("VHDL_Entity was parsed twice?!? filename:" + filename);
			rv = false;
		}
		else
		{			
			try
			{
				BufferedReader in = new BufferedReader(new FileReader(filename));
				final String stageNames[] = {"STAGE_BEFORE_IO", "STAGE_BEFORE_INTERNAL",
						"STAGE_BEFORE_STRUCTURAL", "STAGE_ENDING"};
				
				// these tokens must be located in order before we detect the end of each stage
				String[][] locationTokens = 
				{
					{"entity", "port", "("}, // STAGE_BEFORE_IO
					{"architecture", "is"}, // STAGE_BEFORE_INTERNAL
					{"begin"},  // STAGE_BEFORE_STRUCTURAL
					{} // STAGE_ENDING
				};
				
				// can't create arrays of generics? hmm java :( 
				ArrayList <ArrayList<String> > lineStorage = new ArrayList <ArrayList<String> >();
				lineStorage.add(beforeIOSignals);
				lineStorage.add(beforeInternalSignals);
				lineStorage.add(beforeStructuralInstances);
				lineStorage.add(rest);
	 			
				int locationIndex = 0; // the index we're currently on from locationTokens
				int stage = STAGE_BEFORE_IO; // the current stage
				boolean addingEntityTokens = false;
				boolean addingGenericTokens = false;
				boolean foundGeneric = false; // found the 'generic' token before io?
				int numOpenPar = 0; // the number of open parenthesis, for parsing io/generic signals
				boolean nextName = false; // is the next token the name of the entity?
				
				// for each line
				for (String line = in.readLine(); line != null; line = in.readLine())
				{		
					VHDL_Tokenizer v = new VHDL_Tokenizer(line);
					ArrayList <String> tokens = v.getTokens();
					ArrayList <String> store = lineStorage.get(stage);
					boolean addLine = true; // should add the line to store (used to suppress adding)
					
					// check for the module name
					if (stage == STAGE_BEFORE_IO)
					{
						for (int x = 0; x < tokens.size(); ++x)
						{
							String t = tokens.get(x);
							
							// note this may match more than once if there's two 'entity' tokens
							if (t.equals("entity"))
								nextName = true;
							else if (nextName == true)
							{
								nextName = false;
								name = t;
							}
						}
					}
					
					// pump the IO signal tokens
					if (addingEntityTokens)
					{
						for (int x = 0; x < tokens.size(); ++x)
						{
							String t = tokens.get(x);
							
							if (t.equals(")") && numOpenPar == 0)
							{
									pumpSignalToken(TYPE_IO, ";");
									
									addingEntityTokens = false;
									break;
							}
							else
							{
								if (t.equals(")"))
									--numOpenPar;
								else if (t.equals("("))
									++numOpenPar;
									
								pumpSignalToken(TYPE_IO, t);
							}
						}
					}
					
					if (addingGenericTokens)
					{
						for (int x = 0; x < tokens.size(); ++x)
						{
							String t = tokens.get(x);
							
							if (t.equals(")") && numOpenPar == 0)
							{
								pumpSignalToken(TYPE_GENERIC, ";");
								
								// stop adding generic signal tokens
								addingGenericTokens = false;
								foundGeneric = false;
								break;
							}
							else
							{
								if (t.equals(")"))
									--numOpenPar;
								else if (t.equals("("))
									++numOpenPar;
								
								pumpSignalToken(TYPE_GENERIC, t);
							}
						}
					}
					
					// check if we should be adding generic tokens
					if (stage == STAGE_BEFORE_IO && !addingGenericTokens)
					{
						for (int x = 0; x < tokens.size(); ++x)
						{
							String t = tokens.get(x).toLowerCase();
							
							if (!foundGeneric && t.equals("generic"))
								foundGeneric = true;
							else if (!addingGenericTokens && foundGeneric)
							{
								if (t.equals("("))
									addingGenericTokens = true;
								else
									System.out.println("Parse Error, generic not followed up by '('");
							}
							else if (addingGenericTokens)
							{
								pumpSignalToken(TYPE_GENERIC, t);
							}						
						}
					}
	 				
					if (locationTokens[stage].length == 0) // if we're at the ending
						store.add(line);
					else
					{					
						// see if we need to split the current line
						// note this doesn't guarantee in-order tokens
						for (int i = tokens.indexOf(locationTokens[stage][locationIndex]);
								 i != -1;
								 i = tokens.indexOf(locationTokens[stage][locationIndex]))
						{
							
							++locationIndex;
							
							// we found the last token of this stage on the current line
							if (locationIndex == locationTokens[stage].length)
							{
								if (i == tokens.size() - 1) // this was the last token on the line
								{
									store.add(line);
									addLine = false; // don't add this line to the next stage
									++stage;
									locationIndex = 0;
									
									// start pumping the io signals
									if (stage == STAGE_BEFORE_INTERNAL)
										addingEntityTokens = true;
									
									break;
								}
								else // we need to split this line in two
								{
									store.add(createLine(tokens,0,i));
									
									// note this may miss tokens if they're in the split part
									lineStorage.get(stage + 1).add(createLine(tokens,i + 1, tokens.size() - 1));
									
									addLine = false;
									++stage;		
									locationIndex = 0;
									
									// pump the io signals
									if (stage == STAGE_BEFORE_INTERNAL)
									{
										// note this won't work if the module both starts and ends on the same 
										// line
										for (int x = i+1; x < tokens.size(); ++x)
											pumpSignalToken(TYPE_IO, tokens.get(x));
										
										addingEntityTokens = true;
									}
									
									break;
								}
							}						
						}
						
						if (addLine)
							store.add(line);
					}				
				}
				
				if (stage != STAGE_ENDING)
				{
					System.out.println("Error parsing file '" + filename + "': Final parsing stage(" + 
							stageNames[stage] + " != STAGE_ENDING");
					rv = false;
				}
	
				in.close();
			}
			catch (IOException e)
			{
				System.out.println("Exception during parse: " + e);
				rv = false;
			}
			
			if (rv)
				parsed = true;
		
		}
			
		return rv;
	}
	
	/**
	 * Pump this token as a generic/io signal; populates public entitySignals arraylist
	 * @param type TYPE_GENERIC or TYPE_IO, depending on what we're parsing
	 * @param t the current token in the entity declaration
	 */
	private void pumpSignalToken(int type, String t)
	{
		ArrayList <VHDL_Signal> store;
		
		if (type == TYPE_IO)
			store = entitySignals;
		else // type == TYPE_GENERIC
			store = genericSignals;
		
		// state is maintained in global variable entityIO_SignalsState
		// signal is maintained in currentIO_Signal
		t = t.toLowerCase();
		
		if (entity_SignalsState == STATE_NAME)
		{
			current_Signal = new VHDL_Signal();
			current_Signal.name = t;
			entity_SignalsState = STATE_COLON;
		}
		else if (entity_SignalsState == STATE_COLON)
		{
			if (!t.equals(":"))
				System.out.println("Parse error in Entity's IO Signal Description(" + current_Signal.name +
					"); colon not present!");
			
			if (type == TYPE_IO)
				entity_SignalsState = STATE_DIRECTION;
			else // type == TYPE_GENERIC
				entity_SignalsState = STATE_TYPE;
		}
		else if (entity_SignalsState == STATE_DIRECTION)
		{
			int nameIndex = -1;
			
			// note this won't catch if they use "internal" as the direction
			for (int x = 0; x < VHDL_Signal.DIRECTION_NAMES.length; ++x)
			{
				if (VHDL_Signal.DIRECTION_NAMES[x].equals(t))
				{
					nameIndex = x;
					break;
				}
			}
			
			if (nameIndex != -1)
			{
				current_Signal.direction = VHDL_Signal.DIRECTION_INDICIES[nameIndex];
			}
			else
				System.out.println("Parse error in Entity's IO Signal Description(" + current_Signal.name +
					"); direction is not in/out/inout!");
			
			entity_SignalsState = STATE_TYPE;
		}
		else if (entity_SignalsState == STATE_TYPE)
		{
			if (t.equals(";"))
			{				
				store.add(current_Signal);
				current_Signal = null;
				entity_SignalsState = STATE_NAME;
			}
			else if (t.equals(":="))
			{
				entity_SignalsState = STATE_INITIAL;
			}
			else
			{ 				
				// add token to the type				
				if (current_Signal.type == null)
					current_Signal.type = t;
				else
					current_Signal.type = current_Signal.type + " " + t;
			}
		}
		else if (entity_SignalsState == STATE_INITIAL)
		{
			if (t.equals(";"))
			{				
				store.add(current_Signal);
				current_Signal = null;
				entity_SignalsState = STATE_NAME;
			}
			else
			{ 				
				// add token to initial				
				if (current_Signal.initial == null)
					current_Signal.initial = t;
				else
					current_Signal.initial = current_Signal.initial + " " + t;
			}
		}
	}
	
	/**
	 * Create a space-separated String from an array of tokens and an index range
	 * @param tokens the tokens we are about
	 * @param startIndex the start index to make the string from in the array, inclusive
	 * @param endIndex the end index to make the string from in the array, inclusive
	 * @return the space-separated String
	 */
	private String createLine(ArrayList<String> tokens, int startIndex, int endIndex) 
	{
		String rv = "";

		for (int i = startIndex; i <= endIndex; ++i)
		{
			rv += tokens.get(i) + " ";
		}
		
		return rv;
	}

	public void addSignal(VHDL_Signal signal)
	{
		toAddSignals.add(signal);
	}
	
	public void addInstance(String instanceName, VHDL_Entity type, ArrayList <String> signalMapping)
	{
		if (!type.parsed)
			System.out.println("addInstance error: type " + type.name + " has not been parsed() correctly!");
		else if (type.entitySignals.size() != signalMapping.size())
			System.out.println("addInstance error: mapping signal count(" + signalMapping.size() 
					+ ") is not the same as entity's(" + type.entitySignals.size() + ")");
		else
		{
			toAddInstanceNames.add(instanceName);
			toAddTypes.add(type);
			toAddMappings.add(signalMapping);
		}
	}
	
	/**
	 * Add behavioral code from a file
	 * @param filename the filename containing the behavioral code
	 */
	public void addBehavior(String filename)
	{
		try
		{
			BufferedReader in = new BufferedReader(new FileReader(filename));
			
			for (String line = in.readLine(); line != null; line = in.readLine())
				toAddBehavior.add(line);
			
			in.close();
		}
		catch (IOException e)
		{
			System.out.println("IOException: " + e);
		}
	}
	
	/**
	 * print out the module to a file
	 * @param filename the filename to use
	 */
	public void printEntity(String filename)
	{
		try
		{
		PrintStream fout = new PrintStream(new BufferedOutputStream(
				new FileOutputStream(filename)));
		
		printEntity(fout);
		}
		catch (IOException e)
		{
			System.out.println(e);
		}
	}
	
	/**
	 * Print out the module to the PrintStream
	 * @param out the PrintStream to write to
	 */
	public void printEntity(PrintStream out)
	{
		// can't create arrays of generics? hmm java :( 
		ArrayList <ArrayList<String> > lineStorage = new ArrayList <ArrayList<String> >();
		lineStorage.add(beforeIOSignals);
		lineStorage.add(beforeInternalSignals);
		lineStorage.add(beforeStructuralInstances);
		lineStorage.add(rest);
		
		for (int stage = STAGE_BEFORE_IO; stage <= STAGE_ENDING; ++stage)
		{						
			for (String s : lineStorage.get(stage))
				out.println(s);
			
			if (stage == STAGE_BEFORE_IO)
			{
				// add additional IO signals
				for (int si = 0; si < toAddSignals.size(); ++si)
				{
					VHDL_Signal s = toAddSignals.get(si);					
					
					if (s.direction != SignalDirection.INTERNAL)
					{
						boolean needSemiColon = needSemiColon(toAddSignals, si, beforeInternalSignals);
						
						out.println("\t" + s.name + " : " 
								+ VHDL_Signal.DIRECTION_NAMES[s.direction.ordinal()] + " " + s.type + 
								(needSemiColon ? ";" : ""));
					}					
				}
			}
			else if (stage == STAGE_BEFORE_INTERNAL)
			{
				// add component declarations
				for (int i = 0; i < toAddTypes.size(); ++i)
				{
					// first check if we already added the component declaration (multiple instances of same)
					VHDL_Entity entity = toAddTypes.get(i);
					boolean alreadyDeclared = false;
					
					for (int x = 0; x < i; ++x)
					{
						VHDL_Entity prev = toAddTypes.get(x);
						
						if (entity.name.equals(prev.name))
						{
							alreadyDeclared = true;
							break;
						}
					}
					
					if (alreadyDeclared)
						continue;
					
					out.println("component " + entity.name + " is");
					
					if (entity.genericSignals.size() > 0)
					{
						out.println("generic (");
						
						for (int x = 0; x < entity.genericSignals.size(); ++x)
						{
							VHDL_Signal s = entity.genericSignals.get(x);
							out.print("\t" + s.name + " : " + s.type);
							
							if (s.initial != null)
								out.print(" := " + s.initial);
							
							if (x + 1 < entity.genericSignals.size())
								out.println(";");
							else
								out.println();
						}
						
						out.println(");\n");
					}
					
					out.println("port (");
					
					for (int x = 0; x < entity.entitySignals.size(); ++x)
					{
						VHDL_Signal s = entity.entitySignals.get(x);
						out.print("\t" + s.name + " : " + s.direction + " " + s.type);
						
						if (s.initial != null)
							out.print(" := " + s.initial);
						
						if (x + 1 < entity.entitySignals.size())
							out.println(";");
						else
							out.println();
					}
					
					out.println(");");
					
					out.println("end component;");
				}
				
				// add additional internal signals
				for (int si = 0; si < toAddSignals.size(); ++si)
				{
					VHDL_Signal s = toAddSignals.get(si);
					
					if (s.direction == SignalDirection.INTERNAL)
						out.println("\t" + s.name + " : " + s.type + ";");
				}
			}
			else if (stage == STAGE_BEFORE_STRUCTURAL)
			{
				// add component instances
				for (int i = 0; i < toAddTypes.size(); ++i)
				{
					VHDL_Entity entity = toAddTypes.get(i);
					ArrayList <String> ioMapping = toAddMappings.get(i);
					String instanceName = toAddInstanceNames.get(i);
					
					out.println(instanceName + " : " + entity.name);
					
					if (entity.genericSignals.size() > 0)
					{
						out.println("generic map (");
						
						for (int x = 0; x < entity.genericSignals.size(); ++x)
						{
							VHDL_Signal s = entity.genericSignals.get(x);
							
							out.print("\t" + s.name + " => " + s.initial);						
							
							if (x + 1 < entity.genericSignals.size())
								out.println(",");
							else
								out.println();
						}
						
						out.println(")");
					}
					
					out.println("port map (");
					
					for (int x = 0; x < entity.entitySignals.size(); ++x)
					{
						VHDL_Signal s = entity.entitySignals.get(x);
						String mapping = ioMapping.get(x);
						
						out.print("\t" + s.name + " => " + mapping);						
						
						if (x + 1 < entity.entitySignals.size())
							out.println(",");
						else
							out.println();
					}
					
					out.println(");\n");
				}
			}
		}
	}

	/**
	 * Check if a semicolon needs to be added after this IO signal line, to comply with 
	 * proper VHDL syntax
	 * @param signals the arraylist of signals we are adding
	 * @param index the index of the current IO signal we're adding
	 * @param restLines the lines following our signal addition before the internal signals
	 * @return true iff we must add a semicolon after the current IO signal line
	 */
	private boolean needSemiColon(ArrayList<VHDL_Signal> signals, int index,
			ArrayList<String> restLines) 
	{
		// This logic behind this is that a semicolon is necessary if there are non-internal signals
		// remaining or if the next token in restLines is not ')'
		boolean semiColonNecessary = false;
		
		for (int i = index + 1; i < signals.size(); ++i)
		{
			VHDL_Signal s = signals.get(i);
			
			if (s.direction != SignalDirection.INTERNAL)
			{
				semiColonNecessary = true;
				break;
			}
		}
		
		// check if the next token in restLines is not ')'
		if (!semiColonNecessary)
		{
			for (String s : restLines)
			{
				VHDL_Tokenizer v = new VHDL_Tokenizer(s);
				String t = v.nextToken();
				
				if (t != null)
				{
					if (!t.equals(")"))
						semiColonNecessary = true;
					
					break;
				}
			}
		}
		
		return semiColonNecessary;
	}
}
