-- Stan Bak
-- Mux template for network simulator - 12-08

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.numeric_std.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;
use work.bak.all;

entity mux is
	generic
	(
		WIDTH	: integer := 3
	);
	port 
	( 
		clock			: in std_logic;
		req				: in std_logic_vector(WIDTH-1 downto 0);
		write_sig		: in std_logic_vector(WIDTH-1 downto 0);
		write_data		: in std_logic_vector(WIDTH * 48 - 1 downto 0);
		
		grant			: out std_logic_vector(WIDTH-1 downto 0) := (others => '0');
		
		-- propagate to memory
		write_to_mem	: out std_logic := '0';
		data_to_mem		: out std_logic_vector(47 downto 0) := (others => '0')
		
		
	); 
end mux;

architecture Behavioral of mux is

	signal index : integer range 0 to WIDTH-1 := 0;
	
begin

	write_to_mem <= write_sig(index);
	data_to_mem <= write_data(48*index + 47 downto 48*index);

process (clock) 

variable found : std_logic;

begin
	if rising_edge(clock) then
		if req(index) = '0' then
			grant(index) <= '0';
			found := '0';
		
			FOR i IN 0 TO WIDTH-1 loop
				if found = '0' and req(i) = '1' then
					grant(i) <= '1';
					index <= i;		
					found := '1';
				end if;
			END loop;
			
			
		end if;	
	end if;
end process;

end Behavioral;