-- Stan Bak
-- Mux template for network simulator - 12-08

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.numeric_std.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;
use work.bak.all;

entity mux is
	port 
	( 
		clock			: in std_logic;
		req				: in std_logic_vector(2 downto 0);
		write_sig		: in std_logic_vector(2 downto 0);
		write_data		: in std_logic_vector(2 * 48 - 1 downto 0);
		
		grant			: out std_logic_vector(2 downto 0) := (others => '0');
		
		-- propagate to memory
		write_to_mem	: out std_logic := '0';
		data_to_mem		: out std_logic_vector(47 downto 0) := (others => '0')
		
		
	); 
end traffic_gen;

architecture Behavioral of traffic_gen is

	signal index : integer range 0 to 2 := 0;


begin

	grant(index) <= '1';
	write_to_mem <= write_sig(index);
	data_to_mem <= write_data(48*index + 47 downto 48*index);

process (clock) begin
	if rising_edge(clock) then
		if req(index) = '0' then
			
			-- see if we need to assign a new index as the mutex holder
			if req(0) = '1' then
				index <= 0;
			elsif req(1) = '1' then
				index <= 1;
			elsif req(2) = '1' then
				index <= 2;
			end if;
			
		end if;	
	end if;
end process;

end Behavioral;