-- Stan Bak
-- Useful functions for network simulator


library IEEE;
use IEEE.STD_LOGIC_1164.all;
use IEEE.numeric_std.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

package bak is
	-- this is specific to the pci_record_initial project
	
  subtype type_data is std_logic_vector(64 downto 0);
  -- 71 -> stop
  -- 70 -> idsel
  -- 69 -> reset
  -- 68 -> trdy
  -- 67 -> irdy
  -- 66 -> devsel
  -- 65 -> frame
  -- (64 downto 61) -> clock
  -- (57 downto 50) -> cmd
  -- (47 downto 0) -> addr/data

  function unsigned_to_hexstring  (signal num : in unsigned) return std_logic_vector;
  function unsigned_to_hexstring_c  (constant num : in unsigned) return std_logic_vector;
end bak;


package body bak is

-- signal version
function unsigned_to_hexstring (signal num: in unsigned) return std_logic_vector is

variable output: std_logic_vector(num'length*2 - 1 downto 0);
variable digit : unsigned (7 downto 0);
variable hex_digit : std_logic_vector(7 downto 0);
		
begin

	for i in 0 to (num'length/4)-1 loop
		digit := x"0" & unsigned( num((num'length-1)-4*i downto (num'length-4)-4*i) );
					
		if digit < 10 then
			hex_digit :=  std_logic_vector(digit + 48); -- 48 = '0'
		else
			hex_digit :=  std_logic_vector(digit + 87); -- 97 = 'a'
		end if;	
		
		output((output'length-1)-8*i downto (output'length-8)-8*i) := hex_digit;
	end loop;

return output;

end unsigned_to_hexstring;

-- constant version
function unsigned_to_hexstring_c (constant num: in unsigned) return std_logic_vector is

variable output: std_logic_vector(num'length*2 - 1 downto 0);
variable digit : unsigned (7 downto 0);
variable hex_digit : std_logic_vector(7 downto 0);
		
begin

	for i in 0 to (num'length/4)-1 loop
		digit := x"0" & unsigned( num((num'length-1)-4*i downto (num'length-4)-4*i) );
					
		if digit < 10 then
			hex_digit :=  std_logic_vector(unsigned(digit + 48)); -- 48 = '0'
		else
			hex_digit :=  std_logic_vector(unsigned(digit + 87)); -- 97 = 'a'
		end if;	
		
		output((output'length-1)-8*i downto (output'length-8)-8*i) := hex_digit;
	end loop;

return output;

end unsigned_to_hexstring_c;
 
end bak;
